use std::{
    collections::HashMap,
    error::Error,
    net::TcpStream,
    vec::IntoIter,
};

use imap::Session;
use rustls_connector::RustlsConnector;
use crate::message::Message;
use crate::message_reader::EmailReader;

pub struct ImapReader {
    host: String,
    port: u16,
    username: String,
    password: String,
}

impl ImapReader {
    pub fn new(host: String, port: u16, username: String, password: String) -> Self {
        ImapReader {
            host,
            port,
            username,
            password,
        }
    }

    fn connect(&self) -> Result<HashMap<String, Vec<u8>>, Box<dyn Error>> {
        let mut session = self.open_session()?;

        session.examine("INBOX")?;

        let items = match session.uid_search("ALL") {
            Ok(i) => i,
            Err(e) => return Err(Box::new(e)),
        };

        let mut msgs = HashMap::<String, Vec<u8>>::with_capacity(items.len());

        //println!("# of messages: {}", &items.len());

        for item in items {
            let msg = session.uid_fetch(&item.to_string(), "(BODY.PEEK[] UID)")?;

            let message = if let Some(m) = msg.iter().next() {
                m
            } else {
                continue;
            };

            let body = message.body().expect("Message did not have a body.");

            msgs.insert(item.to_string(), body.to_owned());
        }

        session.logout().expect("Could not log out");

        Ok(msgs)
    }

    fn open_session(
        &self,
    ) -> Result<
        Session<
            rustls_connector::rustls::StreamOwned<
                rustls_connector::rustls::ClientConnection,
                TcpStream,
            >,
        >,
        Box<dyn Error + 'static>,
    > {
        let stream = TcpStream::connect((self.host.as_ref(), self.port))?;
        let tls = RustlsConnector::new_with_webpki_roots_certs();
        let tls_stream = tls.connect(&self.host, stream)?;

        let client = imap::Client::new(tls_stream);

        Ok(client
            .login(&self.username, &self.password)
            .map_err(|e| e.0)?)
    }
}

impl EmailReader for ImapReader {
    fn read_rfc822_messages(&mut self) -> Box<IntoIter<Message>> {
        let msgs = match self.connect() {
            Ok(m) => m,
            Err(e) => {
                dbg!(e);
                return Box::new(Vec::new().into_iter());
            }
        };

        let items = msgs
            .iter()
            .map(|i| Message::new(i.0.to_owned(), i.1.to_owned()));

        let iter = items.collect::<Vec<Message>>().into_iter();

        Box::new(iter)
    }
}
